
#include <stdio.h>
#include <stdlib.h>
#include "tiffio.h"
#include "cmimage.h"
#include "readtiff.h"

#define ERROR_NO	11
#define ERROR_SL	1
#define ERROR_SIZE	2
#define ERROR_IMAGE	3
#define ERROR_NO_PI	4
#define ERROR_PI	5
#define ERROR_BPS	6
#define ERROR_SPP	7
#define ERROR_PC	8
#define ERROR_NO_CM	9
#define ERROR_CM	10

static char	*Error[ERROR_NO]={
		"read error.",
/* SL    */	"scanline data allocation error.",
/* SIZE  */	"no image size description.",
/* IMAGE */	"image data allocation error.",
/* NO_PI */	"no photometric interpretation.",
/* PI    */	"unacceptable photometric interpretation.",
/* BPS   */	"unacceptable bits per sample.",
/* SPP   */	"unacceptable samples per pixel.",
/* PC    */	"unacceptable planar configuation.",
/* NO_CM */	"no color map.",
/* CM    */	"color map data allocation error."
		};

char	*ReadTIFFError(code)
int	code;
{
	if (code<1 || code>=ERROR_NO) return(*Error); return(Error[code]);
}

#define TRS	TIFFReadScanline
#define XSIZE	image->xSize
#define YSIZE	image->ySize
#define R	image->r
#define G	image->g
#define B	image->b

#define UNPACK(bps,ptr,bit,pxl)	{\
	int idx; for (pxl=idx=0; idx<bps; idx++) {\
		     pxl<<=1; if (bit&(*ptr)) pxl|=1;\
		     if ((bit>>=1)==0) { ++ptr; bit=0x80; }\
		 }\
	}

static void	RGBContigImage(tiff,image,tsl,bps,spp)
TIFF		*tiff;
Image		*image;
uint8		*tsl;
uint16		bps,spp;
{
	int	y,x,pxl;
	uint8	*ptr,bit;

	for (y=0; y<YSIZE; y++)
	    if (TRS(tiff,ptr=tsl,y,(uint16)0)>=0)
		if (bps==8)
		    for (x=0; x<XSIZE; x++) {
			R[x][y]=(*(ptr++));
			G[x][y]=(*(ptr++));
			B[x][y]=(*(ptr++)); if (spp==4) ++ptr;
		    }
		else
		    for (bit=0x80, x=0; x<XSIZE; x++) {
			UNPACK(bps,ptr,bit,pxl); R[x][y]=(byte)pxl;
			UNPACK(bps,ptr,bit,pxl); G[x][y]=(byte)pxl;
			UNPACK(bps,ptr,bit,pxl); B[x][y]=(byte)pxl;
			if (spp==4)
			    for (pxl=0; pxl<bps; pxl++)
				if ((bit>>=1)==0) { ++ptr; bit=0x80; }
		    }
}

static void	RGBSeparateImage(tiff,image,tsl,bps)
TIFF		*tiff;
Image		*image;
uint8		*tsl;
uint16		bps;
{
	uint16	spl;
	int	y,x,pxl;
	byte	**rgb;
	uint8	*ptr,bit;

	for (spl=0; spl<3; spl++) {
	    switch(spl) {
		case 0 : rgb=R; break;
		case 1 : rgb=G; break;
		case 2 : rgb=B; break;
	    }
	    for (y=0; y<YSIZE; y++)
		if (TRS(tiff,ptr=tsl,y,spl)>=0)
		    if (bps==8)
			for (x=0; x<XSIZE; x++) rgb[x][y]=(*(ptr++));
		    else
			for (bit=0x80, x=0; x<XSIZE; x++) {
			    UNPACK(bps,ptr,bit,pxl); rgb[x][y]=(byte)pxl;
			}
	}
}

static void	CMImage(tiff,image,tsl,bps)
TIFF		*tiff;
Image		*image;
uint8		*tsl;
uint16		bps;
{
	int	y,x,pxl;
	uint8	*ptr,bit;
	uint16	*ptr16;

	for (y=0; y<YSIZE; y++)
	    if (TRS(tiff,tsl,y,(uint16)0)>=0)
		if (bps==8)
		    for (ptr=tsl, x=0; x<XSIZE; x++) {
			pxl=(int)(*(ptr++)); PUTPIXEL(image,x,y,pxl);
		    }
		else if (bps==16)
		    for (ptr16=(uint16 *)tsl, x=0; x<XSIZE; x++) {
			pxl=(int)(*(ptr16++)); PUTPIXEL(image,x,y,pxl);
		    }
		else
		    for (ptr=tsl, bit=0x80, x=0; x<XSIZE; x++) {
			UNPACK(bps,ptr,bit,pxl); PUTPIXEL(image,x,y,pxl);
		}
}

#define TGF	TIFFGetField
#define TGFD	TIFFGetFieldDefaulted
#define PIXELS	(*cm)->pixels
#define CMR	(*cm)->r
#define CMG	(*cm)->g
#define CMB	(*cm)->b

int	ReadTIFF(tiff,image,cm)
TIFF	*tiff;
Image	**image;
CM	**cm;
{
	uint8	*tsl=(uint8 *)malloc(TIFFScanlineSize(tiff));
	uint32	iw,il;
	uint16	pi,bps,spp,pc,*r,*g,*b;
	int	pxl,idx;

	if (tsl==NULL) return(ERROR_SL);

	if (!TGF(tiff,TIFFTAG_IMAGEWIDTH, &iw) ||
	    !TGF(tiff,TIFFTAG_IMAGELENGTH,&il)) return(ERROR_SIZE);

	if ((*image=AllocImage((int)iw,(int)il))==NULL) return(ERROR_IMAGE);

	if (!TGF(tiff,TIFFTAG_PHOTOMETRIC,&pi)) return(ERROR_NO_PI);

	if (!TGF(tiff,TIFFTAG_BITSPERSAMPLE,&bps))
	    TGFD(tiff,TIFFTAG_BITSPERSAMPLE,&bps);

	if (!TGF(tiff,TIFFTAG_SAMPLESPERPIXEL,&spp))
	    TGFD(tiff,TIFFTAG_SAMPLESPERPIXEL,&spp);

	if (!TGF(tiff,TIFFTAG_PLANARCONFIG,&pc))
	    TGFD(tiff,TIFFTAG_PLANARCONFIG,&pc);

	switch(pi) {
	    case PHOTOMETRIC_RGB :
		if (bps>8) return(ERROR_BPS);
		if (spp!=3 && spp!=4) return(ERROR_SPP);

		*cm=NULL;
		switch(pc) {
		    case PLANARCONFIG_CONTIG :
			RGBContigImage(tiff,*image,tsl,bps,spp); break;
		    case PLANARCONFIG_SEPARATE :
			RGBSeparateImage(tiff,*image,tsl,bps); break;
		    default :
			return(ERROR_PC);
		}
		break;

	    case PHOTOMETRIC_MINISWHITE :
		if (bps>24) return(ERROR_BPS);
		if (spp!=1) return(ERROR_SPP);
		if (pc!=PLANARCONFIG_CONTIG) return(ERROR_PC);

		if ((*cm=AllocCM(1<<bps))==NULL) return(ERROR_CM);

		pxl=PIXELS-1;
		for (idx=0; idx<=pxl; idx++)
		    CMR[idx]=CMG[idx]=CMB[idx]=(word)(0xffff-(0xffff*idx)/pxl);

		CMImage(tiff,*image,tsl,bps);
		break;

	    case PHOTOMETRIC_MINISBLACK :
	    case PHOTOMETRIC_MASK :
		if (bps>24) return(ERROR_BPS);
		if (spp!=1) return(ERROR_SPP);
		if (pc!=PLANARCONFIG_CONTIG) return(ERROR_PC);

		if ((*cm=AllocCM(1<<bps))==NULL) return(ERROR_CM);

		pxl=PIXELS-1;
		for (idx=0; idx<=pxl; idx++)
		    CMR[idx]=CMG[idx]=CMB[idx]=(word)((0xffff*idx)/pxl);

		CMImage(tiff,*image,tsl,bps);
		break;

	    case PHOTOMETRIC_PALETTE :
		if (bps>24) return(ERROR_BPS);
		if (spp!=1) return(ERROR_SPP);
		if (pc!=PLANARCONFIG_CONTIG) return(ERROR_PC);

		if ((*cm=AllocCM(1<<bps))==NULL) return(ERROR_CM);

		if (!TGF(tiff,TIFFTAG_COLORMAP,&r,&g,&b)) return(ERROR_NO_CM);

		for (idx=0; idx<PIXELS; idx++) {
		    CMR[idx]=r[idx]; CMG[idx]=g[idx]; CMB[idx]=b[idx];
		}

		CMImage(tiff,*image,tsl,bps);
		break;

	    default :
		return(ERROR_PI);
	}
	free(tsl); return(0);
}
