
#include <stdio.h>
#include <stdlib.h>
#include "tiffio.h"
#include "cmimage.h"
#include "readtiff.h"
#include "comp.h"

#define BPS	8

#define GIF_HEADER_LENGTH		6
#define GIF87A_HEADER			"GIF87a"
#define GIF89A_HEADER			"GIF89a"
#define EXTENSION_BLOCK			0x21
#define COMMENT_BLOCK			0xfe	/* GIF89a */
#define IMAGE_SEPARATOR			0x2c
#define GIF_TRAILER			0x3b

#define LEFT_OFFSET			0
#define TOP_OFFSET			0
#define RIGHT_OFFSET			0
#define BOTTOM_OFFSET			0
#define ASPECT_RATIO			0	/* GIF89a */
#define INTERLACE_FLAG			0
#define COLORMAP_RESOLUTION		8
#define GLOBAL_COLORMAP_FLAG		1
#define GLOBAL_COLORMAP_SORT_FLAG	0	/* GIF89a */
#define LOCAL_COLORMAP_FLAG		(1-GLOBAL_COLORMAP_FLAG)
#define LOCAL_COLORMAP_SORT_FLAG	0	/* GIF89a */
#define BACKGROUND_INDEX		0

static int	xSize,ySize,curX,curY;
static byte	**imageData;

static int	GetByte()
{
	if (curX==xSize) { curX=0; ++curY; }

	if (curY==ySize) return(-1);

	return((int)imageData[curX++][curY]);
}

static FILE	*outFile;
static int	outCnt=0;
static byte	outData[255];

static void	FlushOutput()
{
	if (outCnt>0) {
	    fputc(outCnt,outFile); fwrite(outData,1,outCnt,outFile); outCnt=0;
	}
}

static void	PutByte(data)
int		data;
{
	outData[outCnt]=(byte)data; if (++outCnt>=255) FlushOutput();
}

static void	PutWord(data,file)
int		data;
FILE		*file;
{
	fputc(data&0xff,file); fputc((data>>8)&0xff,file);
}

#define PUTCOLORMAP(bps,cm,file)	\
{ \
	int	idx,lim=1<<bps; \
\
	for (idx=0; idx<cm->pixels; idx++) { fputc(cm->r[idx],file); \
					     fputc(cm->g[idx],file); \
					     fputc(cm->b[idx],file); } \
	for (; idx<lim; idx++) { fputc(255,file); \
				 fputc(255,file); \
				 fputc(255,file); } \
}

static void	WriteGIF(bps,image,cm,desc,file)
int		bps;
Image		*image;
CM		*cm;
char		*desc;
FILE		*file;
{
	int	len,codeSize,comment=(desc!=NULL && strlen(desc)>0);

	if (cm->pixels>(1<<bps)) for (bps=1; cm->pixels>(1<<bps); bps++);

	fwrite((comment)?GIF89A_HEADER:
			 GIF87A_HEADER,1,GIF_HEADER_LENGTH,file);

	PutWord(image->xSize+LEFT_OFFSET+RIGHT_OFFSET,file);
	PutWord(image->ySize+TOP_OFFSET+BOTTOM_OFFSET,file);

#if	GLOBAL_COLORMAP_FLAG==0
	fputc(((COLORMAP_RESOLUTION-1)<<4)|(bps-1),file);
#else
	fputc((comment)?0x80|((COLORMAP_RESOLUTION-1)<<4)|
			     (GLOBAL_COLORMAP_SORT_FLAG<<3)|(bps-1):
			0x80|((COLORMAP_RESOLUTION-1)<<4)|(bps-1),file);
#endif
	fputc(BACKGROUND_INDEX,file);
	fputc((comment)?ASPECT_RATIO:0,file);

#if	GLOBAL_COLORMAP_FLAG!=0
	PUTCOLORMAP(bps,cm,file);
#endif
	if (comment) {
	    fputc(EXTENSION_BLOCK,file);
	    fputc(COMMENT_BLOCK,file);

	    while ((len=strlen(desc))>0) {
		if (len>255) len=255;

		fputc(len,file); do fputc(*(desc++),file); while (--len>0);
	    }
	    fputc(0,file);
	}

	fputc(IMAGE_SEPARATOR,file);
	PutWord(LEFT_OFFSET,file);
	PutWord(TOP_OFFSET,file);
	PutWord(image->xSize,file);
	PutWord(image->ySize,file);

#if	LOCAL_COLORMAP_FLAG==0
	fputc((INTERLACE_FLAG<<6)|(bps-1),file);
#else
	fputc((comment)?0x80|(INTERLACE_FLAG<<6)|
			     (LOCAL_COLORMAP_SORT_FLAG<<5)|(bps-1):
			0x80|(INTERLACE_FLAG<<6)|(bps-1),file);

	PUTCOLORMAP(bps,cm,file);
#endif
	codeSize=(bps<=1)?2:bps;
	xSize=image->xSize; curX=0;
	ySize=image->ySize; curY=0; imageData=image->r; outFile=file;

	fputc(codeSize,file);
	Compress(codeSize,GetByte,PutByte,1); FlushOutput();
	fputc(0,file);

	fputc(GIF_TRAILER,file);
}

static void	ByteCM(cm)
CM		*cm;
{
	int	idx;

	for (idx=0; idx<cm->pixels; idx++) { cm->r[idx]>>=8;
					     cm->g[idx]>>=8;
					     cm->b[idx]>>=8; }
}

static void	Error(msg)
char		*msg;
{
	fputs(msg,stderr); fputc('\n',stderr); exit(1);
}

#define STRLEN	256

int	main(argc,argv)
int	argc;
char	**argv;
{
	TIFF	*tiff;
	Image	*image;
	CM	*cm;
	int	code,bps;
	char	str[STRLEN],*desc,*tmp;
	FILE	*gif;

	if (argc!=3 && argc!=4) Error("usage : tiff2gif TIFF {bps} GIF");

	if ((tiff=TIFFOpen(argv[1],"r"))==NULL) Error("TIFF file not found.");

	if ((code=ReadTIFF(tiff,&image,&cm))>0) {
	    sprintf(str,"TIFF %s",ReadTIFFError(code)); Error(str);
	}

	if (TIFFGetField(tiff,TIFFTAG_IMAGEDESCRIPTION,&tmp)) {
	    if ((desc=(char *)malloc(strlen(tmp)+1))==NULL)
		Error("memory allocation error.");

	    strcpy(desc,tmp);
	}
	else
	    desc=NULL;

	TIFFClose(tiff);

	if (argc==3)
	    bps=1;
	else
	    if ((bps=atoi(argv[2]))>BPS) bps=BPS;

	if (cm==NULL) {
	    if ((cm=RGB2CM(image))==NULL) Error("error in RGB2CM.");

	    ByteCM(cm);
	}
	else {
	    ByteCM(cm);

	    if (cm->pixels>(1<<bps))
	    if ((cm=CM2CM(image,cm))==NULL) Error("error in CM2CM.");
	}

	if (cm->pixels>(1<<bps))
	if ((cm=ReduceCM(image,cm,1<<BPS))==NULL) Error("error in ReduceCM.");

	if (argv[argc-1][0]=='-')
	    gif=stdout;
	else
	    if ((gif=fopen(argv[argc-1],"wb"))==NULL)
		Error("GIF file not created.");

	WriteGIF(bps,image,cm,desc,gif); if (gif!=stdout) fclose(gif);

	return(0);
}
