
#include <stdio.h>
#include <stdlib.h>
#include "tiffio.h"
#include "cmimage.h"
#include "readtiff.h"

#define TGIF_SCALE	128.0
#define CX		1.0
#define CY		1.0
#define ID		23
#define MX		1.0
#define MY		1.0
#define PIXELS		256

#ifndef	RGB_TXT
#define RGB_TXT	"/usr/lib/X11/rgb.txt"
#endif

#define CODE_MIN	'`'
#define CODE_MAX	'~'
#define CODE_NUM	(CODE_MAX-CODE_MIN+1)

#define TGIF_RGB(rgb)	(int)(1000.0*(double)rgb/255.0+0.5)

#define STR_LEN	256

#define ERROR(msg)	fputs(msg,stderr); fputc('\n',stderr)

static void	Error(msg)
char		*msg;
{
	ERROR(msg); exit(1);
}

typedef struct List {
		word		r,g,b;
		char		*name;
		struct List	*next;
	} List;

static List	*AllocateList(r,g,b,name)
word		r,g,b;
char		*name;
{
	int	len,idx;
	List	*list;

	for (len=0; name[len]!='\0'; len++) ;

	if ((list=(List *)malloc(sizeof(*list)))==NULL ||
	    (list->name=(char *)malloc(len+1))==NULL)
	    Error("memory allocation error for color list.");

	list->r=r; list->g=g; list->b=b;
	for (idx=0; idx<len; idx++) list->name[idx]=(*(name++));
	list->name[len]='\0'; list->next=NULL; return list;
}

static List	*StoreRGB(cm)
CM		*cm;
{
	int	pixels;
	char	name[STR_LEN];
	word	*r=cm->r,*g=cm->g,*b=cm->b;
	List	*top=NULL,**ptr=(&top);

	for (pixels=cm->pixels; pixels>0; pixels--) {
	    sprintf(name,"#%02X%02X%02X",(int)*r,(int)*g,(int)*b);
	    *ptr=AllocateList(*r,*g,*b,name); ptr=(&((*ptr)->next));
	    ++r; ++g; ++b;
	}
	return top;
}

static List	*ReadRGB(rgb_txt,useName)
char		*rgb_txt;
int		useName;
{
	FILE	*file=fopen(rgb_txt,"r");
	char	line[STR_LEN],name[STR_LEN],junk[STR_LEN];
	word	r,g,b;
	List	*list,*top=NULL,**ptr=(&top);

	if (file==NULL) {
	    fprintf(stderr,"%s : ",rgb_txt); Error("rgb.txt file not found.");
	}
	while (fgets(line,STR_LEN,file)!=NULL)
	    if (sscanf(line,"%hu %hu %hu %s %s",&r,&g,&b,name,junk)==4) {
		for (list=top; list!=NULL; list=list->next)
		    if (list->r==r && list->g==g && list->b==b) break;

		if (list==NULL) {
		    if (!useName)
			sprintf(name,"#%02X%02X%02X",(int)r,(int)g,(int)b);

		    *ptr=AllocateList(r,g,b,name); ptr=(&((*ptr)->next));
		}
	    }

	if (top==NULL) Error("no color data found in rgb.txt file.");

	fclose(file); return top;
}

static int	RGBDistance(r,g,b,list)
word		r,g,b;
List		*list;
{
	int	dr=(int)r-(int)list->r,
		dg=(int)g-(int)list->g,
		db=(int)b-(int)list->b; return dr*dr+dg*dg+db*db;
}

static void	SearchNearest(cm,top)
CM		*cm;
List		*top;
{
	int	pixels,nearest,distance;
	word	*r=cm->r,*g=cm->g,*b=cm->b;
	List	*list,*ptr;

	for (pixels=cm->pixels; pixels>0; pixels--) {
	    nearest=RGBDistance(*r,*g,*b,top); list=top;
	    for (ptr=top->next; ptr!=NULL; ptr=ptr->next)
	        if ((distance=RGBDistance(*r,*g,*b,ptr))<nearest) {
		    nearest=distance; list=ptr;
		}

	    *(r++)=list->r; *(g++)=list->g; *(b++)=list->b;
	}
}

#define PUTS(msg)	fputs(msg,stdout)

static void	PutName(r,g,b,list)
word		r,g,b;
List		*list;
{
	for (; list!=NULL; list=list->next)
	    if (r==list->r && g==list->g && b==list->b) {
		PUTS(list->name); break;
	    }
}

static void	PutCode(cpp,pixel)
int		cpp,pixel;
{
	if (cpp>1) PutCode(cpp-1,pixel/CODE_NUM);

	putchar(CODE_MIN+pixel%CODE_NUM);
}

static void	ByteCM(cm)
CM		*cm;
{
	int	pixels;
	word	*r=cm->r,*g=cm->g,*b=cm->b;

	for (pixels=cm->pixels; pixels>0; pixels--, r++, g++, b++) {
	    *r>>=8; *g>>=8; *b>>=8;
	}
}

static double	ScanCorner(arg)
char		*arg;
{
	double	c=atof(arg); return (c<0.0)?-c:c*TGIF_SCALE;
}

#define DEF_1(format,v)		fprintf(stderr,format,v  ); fputc('\n',stderr)
#define DEF_2(format,x,y)	fprintf(stderr,format,x,y); fputc('\n',stderr)

static void	Usage()
{
	ERROR("usage : tiff2tgif {options} TIFF");
	ERROR("\toptions");
	ERROR("\t\t-a x1 y1 x2 y2 : set image area (whole image).");
	ERROR("\t\t-b             : switch to use BW-bitmap if possible.");
	DEF_2("\t\t-c cx cy       : set corner location (%lg, %lg).",CX,CY);
	DEF_1("\t\t-i ID          : set tgif ID (%d).",ID);
	ERROR("\t\t-l             : switch to use landscape mode.");
	DEF_2("\t\t-m mx my       : set magnification ratio (%lg, %lg).",MX,MY);
	ERROR("\t\t-n             : switch to use color name in rgb.txt file.");
	DEF_1("\t\t-p pixels      : set number of pixel values (%d).",PIXELS);
	ERROR("\t\t-r             : switch to use reduced color.");
	exit(1);
}

int	main(argc,argv)
int	argc;
char	**argv;
{
	int	arg,x1,y1,x2,y2,code,nibble,bit,pixel,y,x,cpp,ppc,
		aSW=0,bSW=0,cSW=0,iSW=0,lSW=0,mSW=0,nSW=0,pSW=0,rSW=0,
		id=ID,pixels=PIXELS;
	double	cx=CX*TGIF_SCALE,cy=CY*TGIF_SCALE,mx=MX,my=MY;
	TIFF	*tiff=NULL;
	Image	*image;
	CM	*cm;
	char	str[STR_LEN];
	List	*list;

	for (arg=1; arg<argc; arg++)
	    if (argv[arg][0]=='-')
		switch(argv[arg][1]) {
		    case 'a' : case 'A' :
			if (aSW || arg>=argc-4) Usage();
			x1=atoi(argv[++arg]);
			y1=atoi(argv[++arg]);
			x2=atoi(argv[++arg]);
			y2=atoi(argv[++arg]);
			aSW=1; break;
		    case 'b' : case 'B' :
			if (bSW) Usage();
			bSW=1; break;
		    case 'c' : case 'C' :
			if (cSW || arg>=argc-2) Usage();
			cx=ScanCorner(argv[++arg]);
			cy=ScanCorner(argv[++arg]);
			cSW=1; break;
		    case 'i' : case 'I' :
			if (iSW || arg>=argc-1) Usage();
			if ((id=atoi(argv[++arg]))<=0) Error("bad tgif ID.");
			iSW=1; break;
		    case 'l' : case 'L' :
			if (lSW) Usage();
			lSW=1; break;
		    case 'm' : case 'M' :
			if (mSW || arg>=argc-2) Usage();
			if ((mx=atof(argv[++arg]))==0.0 ||
			    (my=atof(argv[++arg]))==0.0)
			    Error("bad magnification ratio.");
			mSW=1; break;
		    case 'n' : case 'N' :
			if (nSW) Usage();
			nSW=1; break;
		    case 'p' : case 'P' :
			if (pSW || arg>=argc-1) Usage();
			if ((pixels=atoi(argv[++arg]))<8)
			    Error("bad number of pixel values.");
			pSW=1; break;
		    case 'r' : case 'R' :
			if (rSW) Usage();
			rSW=1; break;
		    default :
			Usage();
		}
	    else {
		if (tiff!=NULL) Usage();

		if ((tiff=TIFFOpen(argv[arg],"r"))==NULL)
		    Error("TIFF image data file not found.");
	    }

	if (nSW && rSW) Error("-n and -r options are not compatible.");

	if (tiff==NULL) Usage();

	if ((code=ReadTIFF(tiff,&image,&cm))>0) {
	    sprintf(str,"TIFF %s",ReadTIFFError(code)); Error(str);
	}
	TIFFClose(tiff);

	if (aSW) {
	    if (x1<0 || x2>=image->xSize || x1>x2 ||
		y1<0 || y2>=image->ySize || y1>y2) Error("bad image area.");

	    for (x=x1; x<=x2; x++) {
		image->r[x]=(&(image->r[x][y1]));
		image->g[x]=(&(image->g[x][y1]));
		image->b[x]=(&(image->b[x][y1]));
	    }
	    image->r=(&(image->r[x1]));
	    image->g=(&(image->g[x1]));
	    image->b=(&(image->b[x1]));
	    image->xSize=x2-x1+1;
	    image->ySize=y2-y1+1;
	}
	x1=(int)cx; x2=(int)(cx+((mx<0.0)?-mx:mx*(double)image->xSize)+0.5);
	y1=(int)cy; y2=(int)(cy+((my<0.0)?-my:my*(double)image->ySize)+0.5);

	if (cm==NULL) {
	    if ((cm=RGB2CM(image))==NULL) Error("error in RGB2CM.");

	    ByteCM(cm);
	}
	else {
	    ByteCM(cm);

	    if ((cm=CM2CM(image,cm))==NULL) Error("error in CM2CM (1).");
	}

	if (cm->pixels>pixels)
	if ((cm=ReduceCM(image,cm,pixels))==NULL) Error("error in ReduceCM.");

	if (rSW) {
	    if (bSW) bSW=(cm->pixels<=2);

	    if (!bSW) list=StoreRGB(cm);
	}
	else {
	    list=ReadRGB(RGB_TXT,nSW);
	    SearchNearest(cm,list);
	    if ((cm=CM2CM(image,cm))==NULL) Error("error in CM2CM (2).");

	    if (bSW) bSW=(cm->pixels<=2);
	}

	PUTS(  "state(");
	printf("%d,",lSW);	/* PageStyle */
	PUTS(  "30,");		/* FileVersion */
	PUTS(  "100,"); 	/* PrintMag */
	PUTS(  "0,");		/* OrgX */
	PUTS(  "0,");		/* OrgY */
	PUTS(  "1,");		/* Zoom */
	PUTS(  "4,");		/* EnglishGridSize */
	PUTS(  "1,");		/* Grid */
	PUTS(  "8,");		/* Color */
	PUTS(  "2,");		/* HoriAlign */
	PUTS(  "2,");		/* VertAlign */
	PUTS(  "0,");		/* LineWidth */
	PUTS(  "0,");		/* Spline */
	PUTS(  "0,");		/* LineStyle */
	PUTS(  "0,");		/* ObjFill */
	PUTS(  "1,");		/* PenPat */
	PUTS(  "1,");		/* TextJust */
	PUTS(  "'Helvetica',");	/* FontName */
	PUTS(  "0,");		/* TextStyle */
	PUTS(  "20,");		/* PointSize */
	PUTS(  "0,");		/* Unused */
	PUTS(  "0,");		/* Dash */
	PUTS(  "0,");		/* GridSystem */
	PUTS(  "10,");		/* MetricGridSize */
	PUTS(  "0,");		/* TextVSpace */
	PUTS(  "0,");		/* ZoomIn */
	PUTS(  "1,");		/* GridShown */
	PUTS(  "1,");		/* MoveMode */
	PUTS(  "0,");		/* TextRotate */
	PUTS(  "16,");		/* RCBoxRadius */
	PUTS(  "1,");		/* UseGray */
	PUTS(  "0,");		/* PageLayoutMode */
	PUTS(  "1,");		/* PageArg1 */
	PUTS(  "1,");		/* PageArg2 */
	puts(  "1).");		/* PageLineShownInTileMode */

	puts(  "page(1,\"\").");

	puts(  "group([");

	if (bSW) {
	    PUTS(  "xbm(");
	    PUTS(  "'black',");		/* Color */
	    printf("%d,",x1);		/* X1 */
	    printf("%d,",y1);		/* Y1 */
	    printf("%d,",x2);		/* X2 */
	    printf("%d,",y2);		/* Y2 */
	    PUTS(  "2,");		/* ObjFill */
	    printf("%d,",id);		/* Id */
	    PUTS(  "0,");		/* Rotation */
	    printf("%d,",image->xSize);	/* ImageW */
	    printf("%d,",image->ySize);	/* ImageH */
	    PUTS(  "0,");		/* Rotate */
	    PUTS(  "0,");		/* Flip */
	    PUTS(  "0,");		/* RealType */
	    PUTS(  "0,");		/* LLX */
	    PUTS(  "0,");		/* LLY */
	    PUTS(  "0,");		/* URX */
	    PUTS(  "0,");		/* URY */
	    PUTS(  "0,");		/* NoBitmap */
	    PUTS(  "0,");		/* Locked */
	    puts(  "0,");		/* SaveEPSF */

	    PUTS("    \"\",");		/* Date */
	    puts("\"\",");		/* File */

	    PUTS("    \""); nibble=0;
	    for (y=0; y<image->ySize; y++) {
		code=0; bit=8;
		for (x=0; x<image->xSize; x++) {
		    if (GETPIXEL(image,x,y)) code|=bit;
		    if ((bit>>=1)==0) {
			if (nibble && nibble%64==0) PUTS("\n     "); ++nibble;
			printf("%x",code); code=0; bit=8;
		    }
		}
		if (bit!=8) {
		    if (nibble && nibble%64==0) PUTS("\n     "); ++nibble;
		    printf("%x",code);
		}
	    }
	    puts("\",[\n]),");
	}
	else {
	    for (cpp=1, ppc=CODE_NUM; ppc<cm->pixels; ppc*=CODE_NUM, cpp++) ;

	    PUTS(  "xpm(");
	    PUTS(  "'white',");		/* Color */
	    printf("%d,",x1);		/* X1 */
	    printf("%d,",y1);		/* Y1 */
	    printf("%d,",x2);		/* X2 */
	    printf("%d,",y2);		/* Y2 */
	    PUTS(  "0,");		/* UnUsedObjFill */
	    printf("%d,",cm->pixels);	/* NumColors */
	    printf("%d,",cpp);		/* CharsPerPixel */
	    PUTS(  "0,");		/* FirstPixelIsBg */
	    printf("%d,",id);		/* Id */
	    PUTS(  "0,");		/* Rotation */
	    printf("%d,",image->xSize);	/* ImageW */
	    printf("%d,",image->ySize);	/* ImageH */
	    PUTS(  "0,");		/* Rotate */
	    PUTS(  "0,");		/* Flip */
	    puts(  "0,[");		/* Locked */

	    for (pixel=0; pixel<cm->pixels; pixel++) {
		PUTS("   \""); PutCode(cpp,pixel); PUTS("\", ");

		putchar('"'); PutName(cm->r[pixel],
				      cm->g[pixel],
				      cm->b[pixel],list); PUTS("\", ");

		printf("%d, %d, %d%s\n",TGIF_RGB(cm->r[pixel]),
					TGIF_RGB(cm->g[pixel]),
					TGIF_RGB(cm->b[pixel]),
					(pixel!=cm->pixels-1)?",":"],[");
		}

	    for (y=0; y<image->ySize; y++) {
		PUTS("   \"");
		for (x=0; x<image->xSize; x++) PutCode(cpp,GETPIXEL(image,x,y));
		putchar('"');
		puts((y!=image->ySize-1)?",":"],[\n]),");
	    }
	}

	PUTS(  "box(");
	PUTS(  "'black',");	/* Color */
	printf("%d,",x1);	/* X1 */
	printf("%d,",y1);	/* Y1 */
	printf("%d,",x2);	/* X2 */
	printf("%d,",y2);	/* Y2 */
	PUTS(  "0,");		/* ObjFill */
	PUTS(  "0,");		/* LineWidth */
	PUTS(  "1,");		/* PenPat */
	printf("%d,",id+1);	/* Id */
	PUTS(  "0,");		/* Dash */
	PUTS(  "0,");		/* Rotation */
	puts(  "0,[\n])\n],");	/* Locked */

	printf("%d,",id+2);	/* Id */
	puts(  "0,[\n]).");	/* Locked */

	return 0;
}
